from generic_serial_input import generic_serial_input
import config
import re
from util import parse_int, datetime_update, utc_synchronisation
from sensor_drv import SensorDrv
import omc048, obs

class gmx501(SensorDrv):
    __syslog = obs.system_log('GMX501')
    def __init__(self, serial, conf):
        SensorDrv.__init__(self)
        conf['reference_table'] = 'GMX501'
        self.__gmx_obj = generic_serial_input(serial, conf)
        self.__utc_sync = config.get_config_or_default(conf, 'utc_time_sync', False)
        self.__utc_offset = omc048.RTC().utc_offset()
        self.__time_deviation_in_sec = 3.0
        if self.__utc_sync:
            self.__syslog.info('UTC automatic time synchronisation enabled on {}'.format(conf['id']))
 
    async def read_values(self):
        if await self.__gmx_obj.read_values():
            #Recondition the original sentence to calculate the checksum
            sentence = ''
            for data in self.__gmx_obj.data[:-1]:
                sentence = sentence + data + ','
            checksum = int(self.__gmx_obj.data[-1][1:],16)
            checksum_msg = self.checksum(sentence)
            if checksum_msg == checksum:
                if self.__utc_sync:
                    try:
                        # rework the datetime string from the gmx501 to match the datetime_update functions
                        date = self.__gmx_obj.data[11][:10]
                        date = date.replace('-','')
                        date = date[2:]
                        date = date[4:6] + date[2:4] + date[0:2] 
                        utc = self.__gmx_obj.data[11][11:]
                        utc = utc.replace(':', '')
                        utc = utc.replace('.', '')
                        utc = utc + '00'
                        utc_synchronisation(datetime_update(utc, date), self.__utc_offset[0], self.__utc_offset[1], self.__time_deviation_in_sec)
                    except Exception:
                        self.__syslog.info('Time and date not available for synchronisation')
                #recondition to add gps data in proper fields instead of clogged in one data item
                gps_items = self.obtain_gps_data(self.__gmx_obj.data[10]) 
                self.data = self.__gmx_obj.data[0:10] + [gps_items[0]] + [gps_items[1]] + [gps_items[2]] +  self.__gmx_obj.data[11:]
                #The statement below forces the gmx501 to obtain a GPS position prior to releasing the data to the driver manager
                #Make sure the 'response_timeout' is set long enough to let the sensor obtain its GPS fix.
                if gps_items[0] != None:
                    return True
            else:
                self.__syslog.warn('Checksum mismatch: {} versus {}'.format(checksum_msg, checksum))

    async def detect_parameters(self, tag_number = ''):
        if await self.__gmx_obj.detect_parameters(tag_number):
            #recondition to add gps data in proper fields instead of clogged in one data item
            self.parameters = self.__gmx_obj.parameters[0:10] + [("Latitude", "deg",  "LAT")] + [("Longitude", "deg",  "LON" )] + [("Height", "m",  "HGHT{}".format(tag_number))] +  self.__gmx_obj.parameters[11:]
            return True

    async def flush(self):
        await self.__gmx_obj.flush()

    def checksum(self, data_str):
        value = 0
        for element in data_str:
            value ^= ord(element)
        return value
        
    def obtain_gps_data(self, gps_string):
        if gps_string == '':
            gps_split = [None,None,None]
        else:
            gps_split = gps_string.split(':')
        return gps_split

