import obs
from sensor_drv import SensorDrv
import uasyncio

class nortek_response():
    def __init__(self, command):
        self.command = command
        self.value = b""

class sensor_cfg:
    AQUADOPP_PROFILER_VELOCITY_DATA_ID = 0x21
    AWAC_VELOCITY_ID = 0x20
    AWAC_WAVE_DATA_ID = 0x30
    AWAC_WAVE_DATA_HEADER_ID = 0x31
    AWAC_WAVE_SUV_ID = 0x36
    AWAC_STAGE_DATA_ID = 0x42
    BYTES_PER_CELL = 9
    PROFILER_SIZE_STD = 32
    PROFILER_START_DYNAMIC_CELLS = 30
    AWAC_SIZE_STD = 120
    AWAC_START_DYNAMIC_CELLS = 118

class NortekCP(SensorDrv):
    def __init__(self, serial, conf):
        SensorDrv.__init__(self)
        self.__conversion_table =  {
                                        #0:  ("Minute"           ,"MM"      ,"MIN"      ,4  ,1  ,),         #conversion table for Aquadopp Profiler Velocity data-structure and AWAC velocity profile data-structure
                                        #1:  ("Second"           ,"SS"      ,"SEC"      ,5  ,1  ,),         # parameter, unit, tag, position of byte in binary, lenght of parameter, factor
                                        #2:  ("Day"              ,"DD"      ,"DAY"      ,6  ,1  ,),
                                        #3:  ("Hour"             ,"HH"      ,"HOUR"     ,7  ,1  ,),
                                        #4:  ("Year"             ,"YY"      ,"YEAR"     ,8  ,1  ,),
                                        #5:  ("Month"            ,"MM"      ,"MONTH"    ,9  ,1  ,),
                                        6:  ("Error"            ,"ER"       ,"ERR1"     ,10 ,2  ,1),
                                        7:  ("Analn1"           ,"AN"       ,"ANA1"     ,12 ,2  ,1),
                                        8:  ("Battery"          ,"V"        ,"BAT"      ,14 ,2  ,0.1),
                                        9:  ("SoundSpeed/Analn2",""         ,"SND"      ,16 ,2  ,1),
                                        10: ("Heading"          ,"dg"       ,"HD"       ,18 ,2  ,0.1),
                                        11: ("Pitch"            ,"dg"       ,"PT"       ,20 ,2  ,0.1),
                                        12: ("Roll"             ,"dg"       ,"RL"       ,22 ,2  ,0.1),
                                        13: ("PressureMSB"      ,"dbar"     ,"PRM"      ,24 ,1  ,0.001),
                                        14: ("Status"           ,""         ,"ST1"      ,25 ,1  ,1),
                                        15: ("PressureLSW"      ,"dbar"     ,"PRL"      ,26 ,2  ,0.001),
                                        16: ("Temperature"      ,"C"        ,"TMP1"     ,28 ,2  ,0.01),
                                    }
        self.__sensor_table =       {
                                        sensor_cfg.AQUADOPP_PROFILER_VELOCITY_DATA_ID: (sensor_cfg.PROFILER_SIZE_STD ,sensor_cfg.BYTES_PER_CELL ,sensor_cfg.PROFILER_START_DYNAMIC_CELLS),
                                        sensor_cfg.AWAC_VELOCITY_ID:  (sensor_cfg.AWAC_SIZE_STD ,sensor_cfg.BYTES_PER_CELL ,sensor_cfg.AWAC_START_DYNAMIC_CELLS)
            	                }
        self.__syslog = obs.system_log('NORTEKCP')
        self.startup_time = 15
        self.__identifier = None
        self.__number_of_cells = None
        self.__cel_size = None
        self.__serial = serial
        self.__sreader = uasyncio.StreamReader(serial)
        self.__sentence_buffer = b""

    async def read_values(self):
        result = await self.get_total_message()
        if result != None:
            self.data = self.decode_binary(result)
            return True
        return False
    
    async def get_total_message(self):
        self.__sentence_buffer = b""
        start = obs.millis()
        while True:
            chars_in_buffer = self.__serial.any()
            if chars_in_buffer != 0:
                data = await self.__sreader.read(chars_in_buffer)
                if data:
                    self.__sentence_buffer += data
                    if self.check_header_lenght(self.__sentence_buffer):
                        self.__sentence_buffer = b""                            #Mismatched header or sentence too long, reset sentence
                    elif self.calc_checksum(self.__sentence_buffer):
                        return self.__sentence_buffer
            elif obs.elapsed_millis(start) > 10000:
                return None
            await uasyncio.sleep_ms(5)
        
     
    def check_header_lenght(self, sentence):
        if sentence:
            if sentence[0] != 0xa5:
                return True
            elif len(sentence) > 3 and len(sentence) > (((sentence[3] << 8) + sentence[2]) * 2):
                self.__syslog.info('Length of message exceeded')
                return True
        return False

    async def detect_parameters(self, tag_number = ''):
        if await self.__initialise_sensor():
            self.sort_keys()                      
            for key in self.__keys:
                self.parameters.append((self.__conversion_table[key][0], self.__conversion_table[key][1], self.__conversion_table[key][2] + tag_number))
            return True
        return False

    async def __initialise_sensor(self):
        result = await self.get_total_message()
        if result:  
            self.__identifier = result[1]
            if self.__identifier == sensor_cfg.AQUADOPP_PROFILER_VELOCITY_DATA_ID:
                self.__syslog.info('Aquadopp profiler detected')
            elif self.__identifier == sensor_cfg.AWAC_VELOCITY_ID:
                self.__syslog.info('Aquadopp Awac velo detected')
            else:
                self.__syslog.info('Unknown Aquadopp detected')
            self.__size_in_bytes = ((result[3]<<8) + result[2]) * 2
            self.__cel_size = self.__sensor_table[self.__identifier][1]
            self.__stand_bytes = self.__sensor_table[self.__identifier][0]
            if(self.__size_in_bytes  - self.__stand_bytes) % 9 != 0:
                self.__stand_bytes += 1
            self.__number_of_cells = int((self.__size_in_bytes - self.__stand_bytes) / self.__cel_size)
            self.complete_conversiontable()
            return True
        return False
    
    def complete_conversiontable(self):
        self.sort_keys()
        start_key = self.__keys[-1]+1
        start_key2 = start_key + 3*self.__number_of_cells
        plus_bytes = 0
        plus_bytes2 = self.__number_of_cells*6
        for j in range(1, 4):
            for i in range(1, self.__number_of_cells + 1):
                self.__conversion_table[start_key] = ("V{} b{}".format(i, j),    "mm/s","V{}-{}".format(i, j),      self.__sensor_table[self.__identifier][2]+plus_bytes,2, 0.001)
                self.__conversion_table[start_key2] = ("A{} b{}".format(i, j),   "counts","A{}-{}".format(i, j),    self.__sensor_table[self.__identifier][2]+plus_bytes2,1, 1)
                plus_bytes += 2
                plus_bytes2 += 1
                start_key += 1
                start_key2 += 1

    async def flush(self):
        chars_in_buffer = self.__serial.any()
        if chars_in_buffer != 0:
            await self.__sreader.read(chars_in_buffer)
        await uasyncio.sleep_ms(1)

    def sort_keys(self):
        self.__keys = list(self.__conversion_table.keys())
        self.__keys.sort()

    def append_data(self,sentence,table,data):
        if table[3] == 24 or table[3] == 26:
            pressure = ((sentence[24] << 16) + sentence[27] + sentence[26])*table[5]
            data.append(pressure)
        elif table[4] == 1:
            data.append(sentence[table[3]])
        elif table[4] == 2:
            data.append(((sentence[table[3]+1] << 8) + sentence[table[3]])*table[5])
        return data

    def decode_binary(self, sentence):
        __data = []
        self.sort_keys()
        for key in self.__keys:
            self.append_data(sentence, self.__conversion_table[key], __data)
        return __data

    def calc_checksum(self, sentence):
        hex_count = 0
        for i in range(0, len(sentence)-2, 2):                      #size byte instead of last two bytes
            hex_count += ((sentence[i+1] << 8) + sentence[i])
        hex_count += 0xb58c                                         #add B58C hex
        hex_count_byte = hex_count.to_bytes(3,'little')             #3 bytes safe till 0xffffff / 0xffff = 256 words or 512 bytes
        if hex_count_byte[:2] == sentence[-2:]:
            return True
        return False

